import click
from tqdm import tqdm

from kks.util.testing import TestSource, RunOptions
from kks.util.common import get_solution_directory, test_number_to_name, find_test_pairs, \
    get_matching_suffix, format_file
from kks.util.script import find_script


@click.command(short_help='Generate tests')
@click.option('-o', '--output-only', is_flag=True,
              help='If specified, only solution will run. Useful to generate output for manually created tests')
@click.option('-g', '--generator', type=click.Path(exists=True),
              help='Script, used to generate .in files')
@click.option('-s', '--solution', type=click.Path(exists=True),
              help='Script, used to generate .out files')
@click.option('-t', '--test', 'tests', type=int, multiple=True,
              help='Test number to generate')
@click.option('-r', '--range', 'test_range', type=int, nargs=2,
              help='Tests to generate')
@click.option('-f', '--force', is_flag=True,
              help='Overwrite .in files')
@click.option('-i', '--ignore-exit-code', is_flag=True,
              help='Dont fail on non-zero exit code')
def gen(output_only, generator, solution, tests, test_range, force, ignore_exit_code):
    """
    Generate tests

    \b
    Example usage:
      kks gen --test 17
      kks gen --range 1 100
      kks gen --test 12 --output-only
      kks gen --range 1 50 --force
      kks gen --generator gen.py --solution solve.py
    """

    directory = get_solution_directory()

    options = RunOptions(
        ignore_exit_code=ignore_exit_code
    )

    generator = find_script(directory, 'gen', default=generator, exists=not output_only)
    solution = find_script(directory, 'solve', default=solution)

    with TestSource(generator, solution, options) as test_source:
        test_pairs = find_tests_to_gen(directory, tests, test_range)
        test_pairs = sorted(test_pairs)

        generate_tests(test_source, test_pairs, output_only, force)


def find_tests_to_gen(directory, tests, test_range):
    """
    Возвращает пары файлов, с которыми будем работать

    В gen можно передать конкретные названия тестов, либо промежуток тестов
    1) Если не передано ничего - используем [1; 100]
    2) Объединяем конкретные тесты и отрезок
    3) Пытаемся найти существующие файлы с нужными названиями
    4) Если нету, добавляем
    """
    tests_dir = directory / 'tests'
    tests_dir.mkdir(exist_ok=True)

    if not tests and not test_range:
        test_range = (1, 100)

    test_numbers = list(tests)
    if test_range:
        l, r = sorted(test_range)
        test_numbers += list(range(l, r + 1))

    test_names = [test_number_to_name(number) for number in test_numbers]

    pairs = list(find_test_pairs(tests_dir, test_names))

    used_names = [input_file.stem for input_file, _ in pairs]
    not_used_names = set(test_names) - set(used_names)

    def output_file_for_input_file(input_file):
        return input_file.with_suffix(get_matching_suffix(input_file.suffix))

    result = []

    # Существующие файлы (добавляем выходной, если его нету)
    result += [
        (input_file, output_file or output_file_for_input_file(input_file))
        for input_file, output_file in pairs
    ]

    # Если пары совсем не существует, добавляем
    result += [
        ((tests_dir / test_name).with_suffix('.in'),
         (tests_dir / test_name).with_suffix('.out'))
        for test_name in not_used_names
    ]

    return result


def generate_tests(test_source, test_pairs, output_only, force):
    generated_tests = 0
    t = tqdm(test_pairs, leave=False)
    for input_file, output_file in t:
        if not output_only and input_file.exists() and not force:
            t.clear()
            click.secho(f'Input file {format_file(input_file)} ', fg='yellow', err=True, nl=False)
            click.secho(f'already exists, skipping. Specify -f to overwrite', fg='yellow', err=True)
            continue

        if output_only and not input_file.exists():
            t.clear()
            click.secho(f'Input file {format_file(input_file)} ', fg='red', err=True, nl=False)
            click.secho(f'does not exist, skipping', fg='red', err=True)
            continue

        if output_file and output_file.exists() and not force:
            t.clear()
            click.secho(f'Output file {format_file(output_file)} ', fg='yellow', err=True, nl=False)
            click.secho('already exists, skipping. Specify -f to overwrite', fg='yellow', err=True)
            continue

        t.set_description(f'Generating test {format_file(input_file)}')

        if not output_only:
            with input_file.open('wb') as f:
                if test_source.generate_input(input_file.stem, stdout=f) is None:
                    return

        with input_file.open('rb') as f_in, output_file.open('wb') as f_out:
            if test_source.generate_output(input_file.stem, stdin=f_in, stdout=f_out) is None:
                return

        generated_tests += 1

    click.secho(f'Generated {generated_tests} tests!', fg='green')



import tempfile
from pathlib import Path

import click

from kks.util.common import test_number_to_name
from kks.util.compat import subprocess
from kks.util.script import run_script, needs_compilation, compile_script


class TestSource:
    def __init__(self, generator, solution, options):
        self.generator = generator
        self.solution = solution
        self.options = options

        self.generator_directory = None
        if needs_compilation(self.generator):
            self.generator_directory = tempfile.TemporaryDirectory(prefix='kks-')

        self.solution_directory = None
        if needs_compilation(self.solution):
            self.solution_directory = tempfile.TemporaryDirectory(prefix='kks-')

    def generate_input(self, test, stdout=subprocess.PIPE):
        return run_script(self.generator, [test], stdout=stdout,
                          ignore_exit_code=self.options.ignore_exit_code)

    def generate_output(self, test, stdin=None, stdout=subprocess.PIPE, input=None):
        return run_script(self.solution, [test], stdin=stdin, stdout=stdout, input=input,
                          ignore_exit_code=self.options.ignore_exit_code)

    def __enter__(self):
        if self.generator_directory is not None:
            self.generator_directory.__enter__()
            path = Path(self.generator_directory.name)

            click.secho('Compiling generator... ', fg='green', err=True)
            self.generator = compile_script(path, self.generator, self.options)
            if self.generator is None:
                click.secho('Compilation failed!', fg='red', err=True)
                raise click.Abort()
            click.secho(f'Successfully compiled!', fg='green', err=True)

        if self.solution_directory is not None:
            self.solution_directory.__enter__()
            path = Path(self.solution_directory.name)

            click.secho('Compiling solution... ', fg='green', err=True)
            self.solution = compile_script(path, self.solution, self.options)
            if self.solution is None:
                click.secho('Compilation failed!', fg='red', err=True)
                raise click.Abort()
            click.secho(f'Successfully compiled!', fg='green', err=True)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.generator_directory is not None:
            self.generator_directory.__exit__(exc_type, exc_val, exc_tb)

        if self.solution_directory is not None:
            self.solution_directory.__exit__(exc_type, exc_val, exc_tb)


class VirtualTestSequence:
    def __init__(self, generator, tests):
        self.test_source = generator
        self.tests = tests

    def __iter__(self):
        for test in self.tests:
            name = test_number_to_name(test)
            input_data = self.test_source.generate_input(name).stdout
            output_data = self.test_source.generate_output(name, input=input_data).stdout
            yield Test.from_data(name, input_data, output_data)

    def __len__(self):
        return len(self.tests)


class Test:
    TYPE_STDIN = 'stdin'
    TYPE_FILE = 'file'
    TYPE_DATA = 'data'

    def __init__(  # TODO use dataclass?
        self,
        name, test_type,
        input_file=None, output_file=None,
        input_data=None, output_data=None
    ):
        self.name = name
        self.test_type = test_type
        self.input_file = input_file
        self.output_file = output_file
        self.input_data = input_data
        self.output_data = output_data

    @classmethod
    def from_stdin(cls):
        return cls(name='stdin', test_type=Test.TYPE_STDIN)

    @classmethod
    def from_file(cls, name, input_file, output_file):
        return cls(
            name=name, test_type=Test.TYPE_FILE, input_file=input_file, output_file=output_file
        )

    @classmethod
    def from_data(cls, name, input_data, output_data):
        return cls(
            name=name, test_type=Test.TYPE_DATA, input_data=input_data, output_data=output_data
        )

    def is_file(self):
        return self.test_type == Test.TYPE_FILE

    def is_data(self):
        return self.test_type == Test.TYPE_DATA

    def is_stdin(self):
        return self.test_type == Test.TYPE_STDIN

    def read_input(self):
        if self.test_type == Test.TYPE_FILE:
            with self.input_file.open('rb') as f_in:
                return f_in.read()
        elif self.test_type == Test.TYPE_DATA:
            return self.input_data
        else:
            return None

    def read_output(self):
        if self.test_type == Test.TYPE_FILE:
            with self.output_file.open('rb') as f_out:
                return f_out.read()
        elif self.test_type == Test.TYPE_DATA:
            return self.output_data
        else:
            return None


class RunOptions:
    def __init__(self,
                 continue_on_error=False,
                 ignore_exit_code=False,
                 asan=True,
                 valgrind=False,
                 is_sample=False):
        self.continue_on_error = continue_on_error
        self.ignore_exit_code = ignore_exit_code
        self.asan = asan
        self.valgrind = valgrind
        self.is_sample = is_sample
